@@ -4,41 +4,26 @@ Classes in this module are heavily inspired by the
 [unittest.mock library](https://docs.python.org/3/library/unittest.mock.html).
 """
 from __future__ import annotations
-from dataclasses import dataclass
-from inspect import isclass, iscoroutinefunction
-from typing import get_type_hints, Any, Callable, Dict, Optional, Tuple, Type
+from inspect import isclass, iscoroutinefunction, isfunction, signature
+from functools import partial
+from typing import get_type_hints, Any, Callable, Dict, NamedTuple, Optional
 
-
-@dataclass(frozen=True)
-class SpyCall:
-    """A dataclass representing a call to a spy.
-
-    Attributes:
-        spy_id: Identifier of the spy that made the call
-        args: Arguments list of the call
-        kwargs: Keyword arguments list of the call
-    """
-
-    spy_id: int
-    spy_name: str
-    args: Tuple[Any, ...]
-    kwargs: Dict[str, Any]
-
-    def __str__(self) -> str:
-        """Stringify the call to something human readable.
-
-        `SpyCall(spy_id=42, spy_name="name", args=(1,), kwargs={"foo": False})`
-        would stringify as `"name(1, foo=False)"`
-        """
-        args_list = [repr(arg) for arg in self.args]
-        kwargs_list = [f"{key}={repr(val)}" for key, val in self.kwargs.items()]
-
-        return f"{self.spy_name}({', '.join(args_list + kwargs_list)})"
+from .spy_calls import SpyCall
 
 
 CallHandler = Callable[[SpyCall], Any]
 
 
+class SpyConfig(NamedTuple):
+    """Spy configuration options passed to create_spy."""
+
+    handle_call: CallHandler
+    spec: Optional[Any] = None
+    name: Optional[str] = None
+    module_name: Optional[str] = None
+    is_async: bool = False
+
+
 class BaseSpy:
     """Spy object base class.
 
@@ -51,13 +36,25 @@ class BaseSpy:
         handle_call: CallHandler,
         spec: Optional[Any] = None,
         name: Optional[str] = None,
+        module_name: Optional[str] = None,
     ) -> None:
         """Initialize a BaseSpy from a call handler and an optional spec object."""
-        self._name = name or (spec.__name__ if spec is not None else "spy")
         self._spec = spec
         self._handle_call: CallHandler = handle_call
         self._spy_children: Dict[str, BaseSpy] = {}
 
+        self._name = name or (spec.__name__ if spec is not None else "spy")
+        self._module_name = module_name
+
+        if module_name is None and spec is not None and hasattr(spec, "__module__"):
+            self._module_name = spec.__module__
+
+        # ensure spy can pass inspect.signature checks
+        try:
+            self.__signature__ = signature(spec)  # type: ignore[arg-type]
+        except Exception:
+            pass
+
     @property  # type: ignore[misc]
     def __class__(self) -> Any:
         """Ensure Spy can pass `instanceof` checks."""
@@ -66,19 +63,33 @@ class BaseSpy:
 
         return type(self)
 
+    def __repr__(self) -> str:
+        """Get a helpful string representation of the spy."""
+        name = self._name
+        if self._module_name:
+            name = f"{self._module_name}.{name}"
+
+        return f"<Decoy mock of {name}>" if self._spec else "<Decoy spy function>"
+
     def __getattr__(self, name: str) -> Any:
         """Get a property of the spy.
 
         Lazily constructs child spies, basing them on type hints if available.
         """
+        # do not attempt to mock magic methods
+        if name.startswith("__") and name.endswith("__"):
+            return super().__getattribute__(name)
+
+        # return previously constructed (and cached) child spies
         if name in self._spy_children:
             return self._spy_children[name]
 
         child_spec = None
+        child_is_async = False
 
         if isclass(self._spec):
             try:
-                # NOTE(mc, 2021-01-05): `get_type_hints` may fail at runtime,
+                # NOTE: `get_type_hints` may fail at runtime,
                 # e.g. if a type is subscriptable according to mypy but not
                 # according to Python, `get_type_hints` will raise.
                 # Rather than fail to create a spy with an inscrutable error,
@@ -95,11 +106,22 @@ class BaseSpy:
         if isinstance(child_spec, property):
             hints = get_type_hints(child_spec.fget)
             child_spec = hints.get("return")
+        elif isclass(self._spec) and isfunction(child_spec):
+            # `iscoroutinefunction` does not work for `partial` on Python < 3.8
+            # check before we wrap it
+            child_is_async = iscoroutinefunction(child_spec)
+            # consume the `self` argument of the method to ensure proper
+            # signature reporting by wrapping it in a partial
+            child_spec = partial(child_spec, None)  # type: ignore[arg-type]
 
         spy = create_spy(
-            handle_call=self._handle_call,
-            spec=child_spec,
-            name=f"{self._name}.{name}",
+            config=SpyConfig(
+                handle_call=self._handle_call,
+                spec=child_spec,
+                name=f"{self._name}.{name}",
+                module_name=self._module_name,
+                is_async=child_is_async,
+            ),
         )
 
         self._spy_children[name] = spy
@@ -123,20 +145,21 @@ class AsyncSpy(BaseSpy):
         return self._handle_call(SpyCall(id(self), self._name, args, kwargs))
 
 
-def create_spy(
-    handle_call: CallHandler,
-    spec: Optional[Any] = None,
-    is_async: bool = False,
-    name: Optional[str] = None,
-) -> Any:
+SpyFactory = Callable[[SpyConfig], Any]
+
+
+def create_spy(config: SpyConfig) -> Any:
     """Create a Spy from a spec.
 
     Functions and classes passed to `spec` will be inspected (and have any type
     annotations inspected) to ensure `AsyncSpy`'s are returned where necessary.
     """
-    _SpyCls: Type[BaseSpy] = Spy
+    handle_call, spec, name, module_name, is_async = config
+    _SpyCls = AsyncSpy if iscoroutinefunction(spec) or is_async is True else Spy
 
-    if iscoroutinefunction(spec) or is_async is True:
-        _SpyCls = AsyncSpy
-
-    return _SpyCls(handle_call=handle_call, spec=spec, name=name)
+    return _SpyCls(
+        handle_call=handle_call,
+        spec=spec,
+        name=name,
+        module_name=module_name,
+    )
